from .billing import (
	BillingAgreement, BillingPlan, ChargeModel, PaymentDefinition, PreparedBillingAgreement
)
from .disputes import Dispute
from .payer import Payer
from .payments import Payment, Refund, Sale
from .webhooks import WebhookEvent, WebhookEventTrigger


__all__ = (
	"BillingAgreement", "BillingPlan", "ChargeModel", "Dispute", "Payment",
	"Payer", "PaymentDefinition", "PreparedBillingAgreement", "Refund", "Sale",
	"WebhookEvent", "WebhookEventTrigger"
)



from django.contrib import admin
from django.utils.html import format_html

from . import models
from .settings import PAYPAL_WEBHOOK_ID


class BasePaypalObjectAdmin(admin.ModelAdmin):
	_common_fields = ("id", "djpaypal_created", "djpaypal_updated", "livemode")
	change_form_template = "djpaypal/admin/change_form.html"

	def get_fieldsets(self, request, obj=None):
		# Have to remove the fields from the common set, otherwise they'll show up twice.
		fields = [f for f in self.get_fields(request, obj) if f not in self._common_fields]
		return (
			(None, {"fields": self._common_fields}),
			(self.model.__name__, {"fields": fields}),
		)

	def get_list_display(self, request):
		return ("__str__", ) + self.list_display + self._common_fields[1:]

	def get_list_filter(self, request):
		return self.list_filter + ("livemode", )

	def get_readonly_fields(self, request, obj=None):
		return self.readonly_fields + self._common_fields

	def get_search_fields(self, request):
		return self.search_fields + ("id", )

	def has_add_permission(self, request):
		return False


@admin.register(models.BillingPlan)
class BillingPlanAdmin(BasePaypalObjectAdmin):
	list_display = ("state", "type", "create_time")
	list_filter = ("type", "state", "create_time", "update_time")
	raw_id_fields = ("payment_definitions", )

	def activate_plans(self, request, queryset):
		for obj in queryset:
			obj.activate()

	actions = (activate_plans, )


@admin.register(models.BillingAgreement)
class BillingAgreementAdmin(BasePaypalObjectAdmin):
	list_display = ("user", "state")
	list_filter = ("state", )
	raw_id_fields = ("user", "payer_model")

	def cancel(self, request, queryset):
		for agreement in queryset:
			agreement.cancel(note="Cancelled by admin")

	actions = (cancel, )


@admin.register(models.PreparedBillingAgreement)
class PreparedBillingAgreementAdmin(admin.ModelAdmin):
	list_display = (
		"__str__", "user", "executed_agreement", "executed_at", "created", "updated", "livemode"
	)
	list_filter = ("livemode", "executed_at")
	readonly_fields = ("id", "created", "updated")
	raw_id_fields = ("user", "executed_agreement")

	def has_add_permission(self, request):
		return False


@admin.register(models.ChargeModel)
class ChargeModelAdmin(BasePaypalObjectAdmin):
	list_display = ("type", )
	list_filter = ("type", )


@admin.register(models.Dispute)
class DisputeAdmin(BasePaypalObjectAdmin):
	list_display = ("status", "reason", "create_time")
	list_filter = ("status", "reason")
	ordering = ("-create_time", )
	readonly_fields = (
		"create_time", "update_time", "disputed_transactions", "reason",
		"dispute_amount", "dispute_outcome", "seller_response_due_date", "dispute_flow"
	)


@admin.register(models.Payer)
class PayerAdmin(admin.ModelAdmin):
	list_display = (
		"__str__", "first_name", "last_name", "email", "user", "livemode"
	)
	search_fields = ("id", "first_name", "last_name", "email")
	raw_id_fields = ("user", )

	def has_add_permission(self, request):
		return False


@admin.register(models.PaymentDefinition)
class PaymentDefinitionAdmin(BasePaypalObjectAdmin):
	list_display = (
		"type", "frequency", "frequency_interval", "cycles",
	)
	list_filter = ("type", "frequency")
	raw_id_fields = ("charge_models", )


@admin.register(models.Refund)
class RefundAdmin(BasePaypalObjectAdmin):
	date_hierarchy = "create_time"
	list_display = (
		"state", "invoice_number", "refund_reason_code", "create_time",
	)
	list_filter = ("refund_reason_code", )
	raw_id_fields = ("sale", "parent_payment")
	ordering = ("-create_time", )
	readonly_fields = (
		"state", "sale", "parent_payment", "refund_reason_code",
		"refund_funding_type", "create_time", "update_time",
	)


@admin.register(models.Sale)
class SaleAdmin(BasePaypalObjectAdmin):
	date_hierarchy = "create_time"
	list_display = ("state", "create_time", "update_time")
	list_filter = ("state", "payment_mode")
	raw_id_fields = ("billing_agreement", "parent_payment")
	ordering = ("-create_time", )
	readonly_fields = (
		"amount", "payment_mode", "state", "reason_code",
		"protection_eligibility", "protection_eligibility_type",
		"clearing_time", "transaction_fee", "receivable_amount",
		"exchange_rate", "fmf_details", "receipt_id", "parent_payment",
		"processor_response", "billing_agreement", "soft_descriptor",
		"create_time", "update_time",
	)
	search_fields = ("receipt_id", )


@admin.register(models.WebhookEvent)
class WebhookEventAdmin(BasePaypalObjectAdmin):
	list_display = ("event_type", "resource_type", "resource_id_link", "create_time")
	list_filter = ("create_time", "event_type")
	ordering = ("-create_time", )
	readonly_fields = (
		"summary", "event_type", "resource_type", "resource_id_link", "create_time",
		"event_version", "resource", "status", "transmissions",
	)
	search_fields = ("resource_id", )

	def resource_id_link(self, obj):
		return format_html(
			'<strong><a href="{}">{}</a></strong>',
			obj.get_resource().admin_url,
			obj.resource_id,
		)
	resource_id_link.short_description = "Resource Id"


@admin.register(models.WebhookEventTrigger)
class WebhookEventTriggerAdmin(admin.ModelAdmin):
	list_display = (
		"created", "updated", "valid", "processed", "exception", "webhook_event"
	)
	list_filter = ("created", "valid", "processed")
	raw_id_fields = ("webhook_event", )

	def reverify(self, request, queryset):
		for trigger in queryset:
			if trigger.verify(webhook_id=PAYPAL_WEBHOOK_ID):
				trigger.valid = True
				trigger.save()

	def reprocess(self, request, queryset):
		for trigger in queryset:
			if not trigger.valid:
				# Skip invalid webhooks (never process them)
				continue
			trigger.process()

	def has_add_permission(self, request):
		return False

	actions = (reverify, reprocess)



from decimal import Decimal


def fix_django_headers(meta):
	"""
	Fix this nonsensical API:
	https://docs.djangoproject.com/en/1.11/ref/request-response/
	https://code.djangoproject.com/ticket/20147
	"""
	ret = {}
	for k, v in meta.items():
		if k.startswith("HTTP_"):
			k = k[len("HTTP_"):]
		elif k not in ("CONTENT_LENGTH", "CONTENT_TYPE"):
			# Skip CGI garbage
			continue

		ret[k.lower().replace("_", "-")] = v

	return ret


CURRENCY_SIGILS = {
	"CAD": "$",
	"EUR": "€",
	"GBP": "£",
	"USD": "$",
}


def get_friendly_currency_amount(amount, currency):
	currency = currency.upper()
	sigil = CURRENCY_SIGILS.get(currency, "")
	# Always show two decimal places on the amount
	# Note that amount is usually a string, so convert to Decimal first.
	amount = "{:.2f}".format(Decimal(amount))
	return "{sigil}{amount} {currency}".format(sigil=sigil, amount=amount, currency=currency)


def get_version():
	"""
	Returns the current dj-paypal version
	"""
	from . import __version__
	return __version__



# Generated by Django 2.1.2 on 2019-11-10 15:22

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djpaypal', '0010_payment_hold'),
    ]

    operations = [
        migrations.AddField(
            model_name='dispute',
            name='buyer_response_due_date',
            field=models.DateTimeField(blank=True, editable=False, null=True),
        ),
        migrations.AddField(
            model_name='dispute',
            name='refund_details',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, editable=False, null=True),
        ),
        migrations.AlterField(
            model_name='dispute',
            name='messages',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, editable=False, null=True),
        ),
    ]



# Generated by Django 2.0.2 on 2018-02-06 05:20

from django.db import migrations, models


class Migration(migrations.Migration):

	dependencies = [
		("djpaypal", "0001_initial"),
	]

	operations = [
		migrations.AddField(
			model_name="payer",
			name="time_created",
			field=models.DateTimeField(blank=True, null=True, editable=False),
		),
	]



from django.core import checks

from . import settings as djpaypal_settings


VALID_MODES = ("live", "sandbox")


@checks.register("djpaypal")
def check_paypal_api_key(app_configs=None, **kwargs):
	"""Check that the Paypal API keys are configured correctly"""
	messages = []

	mode = getattr(djpaypal_settings, "PAYPAL_MODE", None)
	if mode not in VALID_MODES:
		msg = "Invalid PAYPAL_MODE specified: {}.".format(repr(mode))
		hint = "PAYPAL_MODE must be one of {}".format(", ".join(repr(k) for k in VALID_MODES))
		messages.append(checks.Critical(msg, hint=hint, id="djpaypal.C001"))

	for setting in "PAYPAL_CLIENT_ID", "PAYPAL_CLIENT_SECRET":
		if not getattr(djpaypal_settings, setting, None):
			msg = "Invalid value specified for {}".format(setting)
			hint = "Add PAYPAL_CLIENT_ID and PAYPAL_CLIENT_SECRET to your settings."
			messages.append(checks.Critical(msg, hint=hint, id="djpaypal.C002"))

	return messages



from django.conf import settings
from django.db import models

from ..fields import JSONField


class Payer(models.Model):
	id = models.CharField(max_length=13, primary_key=True)
	first_name = models.CharField(max_length=64, db_index=True, editable=False)
	last_name = models.CharField(max_length=64, db_index=True, editable=False)
	email = models.CharField(max_length=127, db_index=True, editable=False)
	user = models.ForeignKey(
		settings.AUTH_USER_MODEL, on_delete=models.SET_NULL,
		null=True, blank=True, related_name="paypal_payers",
		help_text="The most recent Django user that transacted as this Payer (best-effort)."
	)
	shipping_address = JSONField(null=True, blank=True, editable=False)
	time_created = models.DateTimeField(null=True, blank=True, editable=False)
	livemode = models.BooleanField()
	djpaypal_created = models.DateTimeField(auto_now_add=True)
	djpaypal_updated = models.DateTimeField(auto_now=True)

	def __str__(self):
		return "{first_name} {last_name} <{email}>".format(
			first_name=self.first_name, last_name=self.last_name, email=self.email
		)



# Generated by Django 2.1.2 on 2018-11-18 22:14

from django.db import migrations
import djpaypal.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djpaypal', '0007_sale_invoice_number'),
    ]

    operations = [
        migrations.AddField(
            model_name='refund',
            name='refund_to_payer',
            field=djpaypal.fields.CurrencyAmountField(editable=False, null=True),
        ),
    ]



import operator
from collections import OrderedDict

from django.utils.translation import ugettext as _


class EnumMetaClass(type):
	@classmethod
	def __prepare__(self, name, bases):
		return OrderedDict()

	def __new__(self, name, bases, classdict):
		members = []
		keys = {}
		choices = OrderedDict()
		for key, value in classdict.items():
			if key.startswith("__"):
				continue
			members.append(key)
			if isinstance(value, tuple):
				value, alias = value
				keys[alias] = key
			else:
				alias = None
			keys[alias or key] = key
			choices[alias or key] = value

		for k, v in keys.items():
			classdict[v] = k

		classdict["__choices__"] = choices
		classdict["__members__"] = members

		# Note: Differences between Python 2.x and Python 3.x force us to
		# explicitly use unicode here, and to explicitly sort the list. In
		# Python 2.x, class members are unordered and so the ordering will
		# vary on different systems based on internal hashing. Without this
		# Django will continually require new no-op migrations.
		classdict["choices"] = tuple(
			(str(k), str(v))
			for k, v in sorted(choices.items(), key=operator.itemgetter(0))
		)

		return type.__new__(self, name, bases, classdict)


class Enum(metaclass=EnumMetaClass):
	pass


class PaypalBool(Enum):
	YES = _("Yes")
	NO = _("No")


class BillingAgreementState(Enum):
	Active = _("Active")
	Cancelled = _("Cancelled")
	Completed = _("Completed")
	Created = _("Created")
	Pending = _("Pending")
	Reactivated = _("Reactivated")
	Suspended = _("Suspended")


class BillingPlanState(Enum):
	CREATED = _("Created")
	ACTIVE = _("Active")
	INACTIVE = _("Inactive")
	DELETED = _("Deleted")


class BillingPlanType(Enum):
	FIXED = _("Fixed")
	INFINITE = _("Infinite")


class ChargeModelType(Enum):
	TAX = _("Tax")
	SHIPPING = _("Shipping")


class DisputeChannel(Enum):
	INTERNAL = _("Internal")
	EXTERNAL = _("External")


class DisputeLifecycleStage(Enum):
	INQUIRY = _("Inquiry")
	CHARGEBACK = _("Chargeback")
	PRE_ARBITRATION = _("Pre-arbitration")
	ARBITRATION = _("Arbitration")


class DisputeFlow(Enum):
	ACH_RETURNS = _("ACH returns")
	ACCOUNT_ISSUES = _("Account issues")
	ADMIN_FRAUD_REVERSAL = _("Admin fraud reversal")
	BILLING = _("Billing")
	CHARGEBACKS = _("Chargebacks")
	COMPLAINT_RESOLUTION = _("Complaint resolution")
	CORRECTION = _("Correction")
	DEBIT_CARD_CHARGEBACK = _("Debit card chargeback")
	FAX_ROUTING = _("Fax routing")
	MIPS_COMPLAINT_ITEM = _("MIPS complaint item")
	MIPS_COMPLAINT = _("MIPS complaint")
	OPS_VERIFICATION_FLOW = _("OPS verification flow")
	PAYPAL_DISPUTE_RESOLUTION = _("Paypal dispute resolution")
	PINLESS_DEBIT_RETURN = _("Pinless debit return")
	PRICING_ADJUSTMENT = _("Pricing adjustment")
	SPOOF_UNAUTH_CHILD = _("Spoof Unauth Child")
	SPOOF_UNAUTH_PARENT = _("Spoof Unauth Parent")
	THIRD_PARTY_CLAIM = _("Third party claim")
	THIRD_PARTY_DISPUTE = _("Third party dispute")
	TICKET_RETRIEVAL = _("Ticket retrieval")
	UK_EXPRESS_RETURNS = _("UK Express returns")
	UNKNOWN_FAXES = _("Unknown faxes")
	VETTING = _("Vetting")
	OTHER = _("Other")


class DisputeReason(Enum):
	MERCHANDISE_OR_SERVICE_NOT_RECEIVED = _("Merchandise or service not received")
	MERCHANDISE_OR_SERVICE_NOT_AS_DESCRIBED = _("Merchandise or service not as described")
	UNAUTHORISED = _("Unauthorized")
	CREDIT_NOT_PROCESSED = _("Credit not processed")
	DUPLICATE_TRANSACTION = _("Duplicate transaction")
	INCORRECT_AMOUNT = _("Incorrect amount")
	PAYMENT_BY_OTHER_MEANS = _("Payment by other means")
	CANCELED_RECURRING_BILLING = _("Canceled recurring billing")
	PROBLEM_WITH_REMITTANCE = _("Problem occurred with the remittance")
	OTHER = _("Other")


class DisputeStatus(Enum):
	OPEN = _("Open")
	WAITING_FOR_BUYER_RESPONSE = _("Waiting for buyer response")
	WAITING_FOR_SELLER_RESPONSE = _("Waiting for seller response")
	UNDER_REVIEW = _("Under review")
	RESOLVED = _("Resolved")
	OTHER = _("Other")


class PaymentDefinitionFrequency(Enum):
	DAY = _("Day")
	WEEK = _("Week")
	MONTH = _("Month")
	YEAR = _("Year")


class PaymentDefinitionType(Enum):
	TRIAL = _("Trial")
	REGULAR = _("Regular")


class PaymentFailureReason(Enum):
	UNABLE_TO_COMPLETE_TRANSACTION = _("Unable to complete transaction")
	INVALID_PAYMENT_METHOD = _("Invalid payment method")
	PAYER_CANNOT_PAY = _("Payer cannot pay")
	CANNOT_PAY_THIS_PAYEE = _("Cannot pay this payee")
	REDIRECT_REQUIRED = _("Redirect required")
	PAYEE_FILTER_RESTRICTIONS = _("Payee filter restrictions")


class PaymentIntent(Enum):
	sale = _("Sale")
	authorize = _("Authorize")
	order = _("Order")


class PaymentState(Enum):
	created = _("Created")
	approved = _("Approved")
	failed = _("Failed")


class TermsType(Enum):
	WEEKLY = _("Weekly")
	MONTHLY = _("Monthly")
	YEARLY = _("Yearly")


class JsonPatchOp(Enum):
	add = _("add")
	remove = _("remove")
	replace = _("replace")
	move = _("move")
	copy = _("copy")
	test = _("test")


class PaypalAction(Enum):
	CONTINUE = _("Continue")
	CANCEL = _("Cancel")


class RefundFundingType(Enum):
	PAYOUT = _("Payout")


class RefundPendingReasonCode(Enum):
	ECHECK = _("eCheck")


class RefundReasonCode(Enum):
	ADJUSTMENT_REVERSAL = _("Adjustment reversal")
	ADMIN_FRAUD_REVERSAL = _("Admin fraud reversal")
	ADMIN_REVERSAL = _("Admin reversal")
	APPEAL = _("Appeal")
	BUYER_COMPLAINT = _("Buyer complaint")
	CHARGEBACK = _("Chargeback")
	CHARGEBACK_SETTLEMENT = _("Chargeback settlement")
	DENIED = _("Denied")
	DISPUTE_PAYOUT = _("Dispute payout")
	FUNDING = _("Funding")
	GUARANTEE = _("Guarantee")
	LIMITS = _("Limits")
	NO_FAULT = _("No fault")
	OTHER = _("Other")
	REFUND = _("Refund")
	REGULATORY_BLOCK = _("Regulatory block")
	REGULATORY_REJECT = _("Regulatory reject")
	REGULATORY_REVIEW_EXCEEDING_SLA = _("Regulatory review exceeding SLA")
	RISK = _("Risk")
	SELLER_FAULT = _("Seller fault")
	SELLER_VOLUNTARY = _("Seller voluntary")
	THIRDPARTY_LOGISTICS_FAULT = _("Third-party logistics fault")
	UNAUTH_CLAIM = _("Unauth claim")
	UNAUTH_SPOOF = _("Unauth spoof")


class RefundState(Enum):
	pending = _("Pending")
	completed = _("Completed")
	cancelled = _("Cancelled")
	failed = _("Failed")


class SalePaymentMode(Enum):
	INSTANT_TRANSFER = _("Instant transfer")
	MANUAL_BANK_TRANSFER = _("Manual bank transfer")
	DELAYED_TRANSFER = _("Delayed transfer")
	ECHECK = _("eCheck")


class SaleState(Enum):
	completed = _("Completed")
	partially_refunded = _("Partially refunded")
	pending = _("Pending")
	refunded = _("Refunded")
	denied = _("Denied")


class SaleReasonCode(Enum):
	CHARGEBACK = _("Chargeback")
	GUARANTEE = _("Guarantee")
	BUYER_COMPLAINT = _("Buyer complaint")
	REFUND = _("Refund")
	UNCONFIRMED_SHIPPING_ADDRESS = _("Unconfirmed shipping address")
	ECHECK = _("eCheck")
	INTERNATIONAL_WITHDRAWAL = _("International withdrawal")
	RECEIVING_PREFERENCE_MANDATES_MANUAL_ACTION = _(
		"Receiving preference mandates manual action"
	)
	PAYMENT_REVIEW = _("Payment review")
	REGULATORY_REVIEW = _("Regulatory review")
	UNILATERAL = _("Unilateral")
	VERIFICATION_REQUIRED = _("Verification required")
	TRANSACTION_APPROVED_AWAITING_FUNDING = _("Transaction approved awaiting funding")


class SaleProtectionEligibility(Enum):
	ELIGIBLE = _("Eligible")
	PARTIALLY_ELIGIBLE = _("Partially eligible")
	INELIGIBLE = _("Ineligible")


class SaleProtectionEligibilityType(Enum):
	ITEM_NOT_RECEIVED_ELIGIBLE = _("Item not received eligible")
	UNAUTHORIZED_PAYMENT_ELIGIBLE = _("Unauthorized payment eligible")
	BOTH = (_("Both"), "ITEM_NOT_RECEIVED_ELIGIBLE,UNAUTHORIZED_PAYMENT_ELIGIBLE")


class SalePaymentHoldStatus(Enum):
	HELD = _("Held")
	RELEASED = _("Released")


class SalePaymentHoldReason(Enum):
	PAYMENT_HOLD = _("Payment hold")
	SHIPPING_RISK_HOLD = _("Shipping risk hold")



from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseServerError
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from .models import WebhookEventTrigger


@method_decorator(csrf_exempt, name="dispatch")
class ProcessWebhookView(View):
	"""
	A Paypal Webhook handler view.

	This will create a WebhookEventTrigger instance, verify it,
	then attempt to process it.

	If the webhook cannot be verified, returns HTTP 400.

	If an exception happens during processing, returns HTTP 500.
	"""
	def post(self, request):
		if "HTTP_PAYPAL_TRANSMISSION_ID" not in request.META:
			# Do not even attempt to process/store the event if there is
			# no paypal transmission id so we avoid overfilling the db.
			return HttpResponseBadRequest()

		trigger = WebhookEventTrigger.from_request(request)

		if trigger.exception:
			# An exception happened, return 500
			return HttpResponseServerError()

		if not trigger.valid:
			# Webhook Event did not validate, return 400
			return HttpResponseBadRequest()

		return HttpResponse(str(trigger.id))



# Generated by Django 2.1.2 on 2018-10-01 11:51

from django.db import migrations

import djpaypal.fields


class Migration(migrations.Migration):

    dependencies = [
        ("djpaypal", "0005_auto_20180919_1447"),
    ]

    operations = [
        migrations.AddField(
            model_name="refund",
            name="refund_from_received_amount",
            field=djpaypal.fields.CurrencyAmountField(
                editable=False,
                help_text=(
                    "The currency and amount of the refund that is subtracted from the "
                    "original payment recipient's PayPal balance."
                ),
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="refund",
            name="refund_from_transaction_fee",
            field=djpaypal.fields.CurrencyAmountField(
                editable=False,
                help_text=(
                    "The currency and amount of the transaction fee that is refunded to "
                    "original the recipient of payment."
                ),
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="refund",
            name="total_refunded_amount",
            field=djpaypal.fields.CurrencyAmountField(
                editable=False,
                help_text=(
                    "The currency and amount of the total refund from the original "
                    "purchase."
                ),
                null=True,
            ),
        ),
    ]
